#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (print_function)

import sys
import logging
import termios
import time
import tty
import re

from collections import namedtuple
from itertools import groupby
from datetime import datetime

from keyriban import config
from keyriban.config import cfg, term


logger = logging.getLogger(__name__)
MODIFIERS = ['CTRL', 'CONTROL', 'COMMAND', 'CMD', 'OPTION', 'ALT', 'ALTGR',
    'SHIFT', 'SHFT']
SPECIALS = ['RETURN', 'UP', 'DOWN', 'LEFT', 'RIGHT', 'DEL', 'ESC']
# Add left and right variants
MODIFIERS += ['L' + x for x in MODIFIERS] + ['R' + x for x in MODIFIERS]
# Regex: 'abc<Ctrl>[LShft]' => ['a', 'b', '<Ctrl>', '[LShft]']
SPLIT_RE = re.compile(r'((?:[<\[](?:.+?)[>\]])|.)', flags=re.I)
Keypress = namedtuple('Keypress', ['key', 'at'])


def start_with_modifier(keys):
    """Return True if first key is a modifier.
    """
    for x in MODIFIERS:
        if '<%s>' % x in keys[0]:
            return True


class LogReader(object):
    """Return keys appended to a log file, in real-time.
    """

    def __init__(self, filepath):
        self.filepath = filepath
        if 'format_closing' not in cfg or 'chars_skip' not in cfg:
            self.detect_log_format()

    def detect_log_format(self):
        """Detect keylogger file lines format.
        """
        term_settings = termios.tcgetattr(sys.stdin)
        tty.setraw(sys.stdin)
        with open(self.filepath, 'r') as keylog:
            keylog.seek(0, 2)
            print(term.clear + '[1/3] Press <Enter>', end='')
            sys.stdin.read(1)
            print(term.clear + '[2/3] Press <Ctrl>+R', end='')
            sys.stdin.read(1)
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, term_settings)
            print(term.clear + '[3/3] Calibration done')
            time.sleep(2)
            line = keylog.readlines()[-1].upper()
        pattern = r'(<(%s)>)R' % '|'.join(MODIFIERS)
        try:
            cfg['chars_skip'] = re.search(pattern, line).start()
        except AttributeError:
            logger.error('Cannot detect key logger line format. ' +
                "Check '%s' content." % self.filepath)
            exit(1)
        pattern = r'(<(%s)>)R<\2>' % '|'.join(MODIFIERS)
        cfg['format_closing'] = re.search(pattern, line) is not None
        config.save()

    def combine_keys(self, keys=None):
        """Generate keys sequece by merging modifier plus following key.
        """
        if not keys:
            keys = []
        keys = [x[0] for x in groupby(keys)]  # remove consecutive duplicates
        if len(self.combo):
            keys = ['+'.join([x.key for x in self.combo] + keys)]
        else:
            keys = [' '.join(keys)]
        if not cfg['format_closing']:
            logger.info('Reset combo')
            self.combo = []
        self.last_key_time = datetime.now()
        return keys

    def combo_outdated(self):
        """Return True if too much time elapsed since modifier pressed.
        """
        res = None
        if self.combo:
            elapsed_time = (datetime.now() - self.combo[0].at).seconds
            res = elapsed_time > 1
        return res

    def tail(self, keep_alive):
        """Yield list of keys as soon as they are logged while preserving
           modifier combos.

           Combos are represented like ['Ctrl+C'] rather than ['Ctrl', 'C'].
           keep_alive is the maximum time elapsed in seconds before yielding an
           empty key [''].
        """
        self.combo = []
        self.last_key_time = datetime.now()
        linefeed = False
        with open(self.filepath, 'r') as logfile:
            logfile.seek(0, 2)  # 2: seek relative to file end
            while True:
                time.sleep(.1)
                line = logfile.readline()
                if linefeed:
                    line = line[cfg['chars_skip'].get(int):]
                linefeed = '\n' in line
                line = line.upper().strip().replace(' ', '<SPACE>')
                keys = filter(None, re.split(SPLIT_RE, line))
                if not keys:
                    # Yield modifier alone when no key pressed during the few
                    # next seconds
                    if self.combo_outdated():
                        keys = self.combine_keys()
                        self.combo = []
                        yield keys
                    # Yield empty string if no key pressed in last seconds
                    if keep_alive and (datetime.now() -
                            self.last_key_time).seconds > keep_alive:
                        keys = self.combine_keys([])
                        yield keys
                    continue
                # Strip closing markup if modifier previously pressed
                if cfg['format_closing'] and (self.combo and
                        [self.combo[0].key] == keys):
                    self.combo = []
                    continue
                # Modifier pressed
                if start_with_modifier(keys):
                    self.combo.extend([Keypress(key=x, at=datetime.now())
                        for x in keys])
                    logger.info('Append to combo %s' % self.combo)
                # Yield keys
                else:
                    keys = self.combine_keys(keys)
                    yield keys
